##############################################################################
#
# Copyright (c) 2001-2003 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Python Object Publisher -- Publish Python objects on web servers

Provide an apply-like facility that works with any mapping object

$Id: publish.py 105429 2009-11-02 07:47:51Z ctheune $
"""
import sys
from zope import component
from zope.interface import implements
from zope.publisher.interfaces import Retry, IReRaiseException
from zope.proxy import removeAllProxies

_marker = object()  # Create a new marker object.

def unwrapMethod(obj):
    """obj -> (unwrapped, wrapperCount)

    Unwrap 'obj' until we get to a real function, counting the number of
    unwrappings.

    Bail if we find a class or something we can't identify as callable.
    """
    wrapperCount = 0
    unwrapped = obj

    for i in range(10):
        bases = getattr(unwrapped, '__bases__', None)
        if bases is not None:
            raise TypeError("mapply() can not call class constructors")

        im_func = getattr(unwrapped, 'im_func', None)
        if im_func is not None:
            unwrapped = im_func
            wrapperCount += 1
        elif getattr(unwrapped, 'func_code', None) is not None:
            break
        else:
            unwrapped = getattr(unwrapped, '__call__' , None)
            if unwrapped is None:
                raise TypeError("mapply() can not call %s" % repr(obj))
    else:
        raise TypeError("couldn't find callable metadata, mapply() error on %s"
                        % repr(obj))

    return unwrapped, wrapperCount

# XXX These imports are for the hacked changes below.
from zope.security.interfaces import ForbiddenAttribute
from zope.security.proxy import Proxy
from zope.security.interfaces import INameBasedChecker
from zope.security.checker import (
    ProxyFactory,
    getChecker,
    getCheckerForInstancesOf,
    _defaultChecker,
    _checkers,
    InterfaceChecker,
    defineChecker,
    undefineChecker)
from zope.proxy import removeAllProxies
import logging

def _describe_checker(checker):
    if INameBasedChecker.providedBy(checker):
        try:
            logging.warning(
                "names and security values for checker "
                "get_permissions: %r",
                checker.get_permissions)
        except:
            logging.warning(
                "Hm, I couldn't look at the get_permissions. Weird.",
                exc_info=True)
    else:
        logging.warning('%r is not a name-based checker.', checker)

def _use_checker(unwrapped, checker_to_use):
    defineChecker(type(unwrapped), checker_to_use)
    new_obj = ProxyFactory(unwrapped)
    if getChecker(new_obj) is not checker_to_use:
        logging.warning(
            'Concerningly, defineChecker did not seem to '
            'work')
        new_obj = ProxyFactory(unwrapped, checker_to_use)
    try:
        # Make sure that the original trigger works now.
        getattr(new_obj, '__call__')
    except ForbiddenAttribute:
        logging.warning(
            'Some combination of the world and I are insane.')
        # Punt!
        new_obj = unwrapped
    return new_obj

_first_was_proxied = None

def emergency_hack_2010_12_17(obj, where_comment):
    # XXX This function and the hacks that call it exist to both
    # workaround and help figure out the problems shown by tracebacks like
    # https://pastebin.canonical.com/36719/ .
    from lazr.restful import CollectionResource
    from lazr.restful.interfaces import ICollectionResource
    global _first_was_proxied
    unwrapped = removeAllProxies(obj)
    unwrapped_type = type(unwrapped)
    # We are suspicious that somehow multiple instances of this class
    # exists, so we check by name, rather than identity.
    if (unwrapped_type.__name__ == 'CollectionResource' and
        unwrapped_type.__module__ == 'lazr.restful._resource'):
        is_proxy = type(obj) is Proxy
        # Record whether the first instance we see is proxied.
        if _first_was_proxied is None:
            _first_was_proxied = is_proxy
            if not is_proxy:
                logging.warning(
                    'Looks like CollectionResources are normally unproxied?')
        # Now start figuring out what to do.
        # The goals of this code are three-fold.
        # 1. Get a new obj that can be called, eliminating the error in
        #    the traceback.  There may be subsequent errors also, but this
        #    papers over the problem we know about.
        # 2. Log more information to verify a hypothesis that the problem
        #    is that we have different instances of the
        #    CollectionResource class, only one of which is registered
        #    with the security machinery.
        # 3. If that hypothesis seems correct, correct for it on the fly.
        #    this will reduce the times that this code is called.
        # This code does not do one important thing: try to determine
        # *how* the interpreter state could have come to the point
        # that we hypothesize.  If the hypothesis is validated and this
        # version of the webservice lives long enough for us to care,
        # we'll investigate that next.  The current best guess is that
        # Grok magic + Django magic are interoperating to cause Grok to
        # be parsed twice, which is doing...something bad yet to be
        # determined.
        if is_proxy and not _first_was_proxied:
            logging.warning(
                'Ripping off proxy %s because it looks like proxies are the '
                'surprise problem.', where_comment)
            obj = unwrapped
        elif is_proxy and getChecker(obj) is _defaultChecker:
            # We believe that this is the problem case.
            logging.warning(
                'Webservice problem encountered %s.', where_comment)
            new_obj = ProxyFactory(unwrapped)
            if getChecker(new_obj) is _defaultChecker:
                checker_by_class = getCheckerForInstancesOf(unwrapped_type)
                if checker_by_class is not None:
                    if checker_by_class is _defaultChecker:
                        logging.warning(
                            'Fixing explicit registration of _defaultChecker')
                        undefineChecker(unwrapped_type)
                        new_obj = _use_checker(
                            unwrapped, InterfaceChecker(ICollectionResource))
                    else:
                        logging.warning(
                            'Even though the we get the default checker from '
                            'ProxyFactory, we get a different checker from '
                            'getCheckerForInstancesOf.  No good known '
                            'action.')
                        _describe_checker(checker_by_class)
                else:
                    # No checker for this class.  Let's see if we can find
                    # a checker for a class like it.
                    checker_to_use = None
                    if unwrapped_type is not CollectionResource:
                        logging.warning(
                            'The identity of the object type is not the same '
                            'as the imported CollectionResource')
                        checker_by_class = getCheckerForInstancesOf(
                            CollectionResource)
                        if checker_by_class is None:
                            logging.warning(
                                'There is no checker for the imported '
                                'version of the class either.')
                            # Is there any checker for a class with this
                            # name, even?
                            for klass, value in _checkers.items():
                                if (klass.__name__ == 'CollectionResource' and
                                    klass.__module__ ==
                                        'lazr.restful._resource'):
                                    logging.warning(
                                        'Another class instance found '
                                        'registered.')
                                    _describe_checker(value)
                                    checker_to_use = value
                        else:
                            logging.warning(
                                'There is a checker for the imported version '
                                'of the class')
                            _describe_checker(checker_by_class)
                            checker_to_use = checker_by_class
                    if checker_to_use is None:
                        logging.warning('Making a checker on the fly.')
                        checker_to_use = InterfaceChecker(ICollectionResource)
                    new_obj = _use_checker(unwrapped, checker_to_use)
            else:
                logging.warning(
                    'Explicitly asking for a proxy of the unwrapped object '
                    'seems to have done the right thing now, even though it '
                    'does not seem to have done so before.')
                _describe_checker(getChecker(new_obj))
            obj = new_obj
    return obj

"""
Test string.  Can be run in Launchpad's bin/py.  Comments show expected
results.

def reset_test():
    global _first_was_proxied
    _first_was_proxied = None
    for k, v in _checkers.items():
        if (k.__name__ == CollectionResource.__name__ and
            k.__module__ == CollectionResource.__module__):
            # remove it.
            undefineChecker(k)

reset_test()
emergency_hack_2010_12_17('foo', 'during test')
# 'foo'
r = CollectionResource.__new__(CollectionResource)
r is emergency_hack_2010_12_17(r, 'during test')
# WARNING:root:Looks like CollectionResources are normally unproxied?
# True
_first_was_proxied
# False
p_r = ProxyFactory(r)
getChecker(p_r) is _defaultChecker
# True
r is emergency_hack_2010_12_17(p_r, 'during test')
# WARNING:root:Ripping off proxy during test because it looks like proxies are the surprise problem.
# True

reset_test()
res = emergency_hack_2010_12_17(p_r, 'during test')
# WARNING:root:Webservice problem encountered during test.
# WARNING:root:Making a checker on the fly.
type(res)
# <type 'zope.security._proxy._Proxy'>
res.__call__ is not None
# True

reset_test()
defineChecker(CollectionResource, _defaultChecker)
res = emergency_hack_2010_12_17(p_r, 'during test')
# WARNING:root:Webservice problem encountered during test.
# WARNING:root:Fixing explicit registration of _defaultChecker.
res.__call__ is not None
# True

reset_test()
class Stub(object):
    def __call__(self):
        pass

Stub.__name__ = CollectionResource.__name__
Stub.__module__ = CollectionResource.__module__
x = Stub()
p_x = ProxyFactory(x)
res = emergency_hack_2010_12_17(p_x, 'during test')
# WARNING:root:Webservice problem encountered during test.
# WARNING:root:The identity of the object type is not the same as the imported CollectionResource
# WARNING:root:There is no checker for the imported version of the class either.
# WARNING:root:Making a checker on the fly.
res.__call__ is not None
# True

reset_test()
defineChecker(CollectionResource, InterfaceChecker(ICollectionResource))
res = emergency_hack_2010_12_17(p_x, 'during test')
# WARNING:root:Webservice problem encountered during test.
# WARNING:root:The identity of the object type is not the same as the imported CollectionResource
# WARNING:root:There is a checker for the imported version of the class
# WARNING:root:names and security values for checker get_permissions: {'__call__': Global(CheckerPublic,zope.security.checker), 'do_GET': Global(CheckerPublic,zope.security.checker), 'getETag': Global(CheckerPublic,zope.security.checker)}
res.__call__ is not None
# True

reset_test()
class AltStub(object):
    def __call__(self):
        pass

AltStub.__name__ = CollectionResource.__name__
AltStub.__module__ = CollectionResource.__module__
defineChecker(AltStub, InterfaceChecker(ICollectionResource))
res = emergency_hack_2010_12_17(p_x, 'during test')
# WARNING:root:Webservice problem encountered during test.
# WARNING:root:The identity of the object type is not the same as the imported CollectionResource
# WARNING:root:There is no checker for the imported version of the class either.
# WARNING:root:Another class instance found registered.
# WARNING:root:names and security values for checker get_permissions: {'__call__': Global(CheckerPublic,zope.security.checker), 'do_GET': Global(CheckerPublic,zope.security.checker), 'getETag': Global(CheckerPublic,zope.security.checker)}
res.__call__ is not None
# True
"""

def mapply(obj, positional=(), request={}):
    __traceback_info__ = obj

    # we need deep access for introspection. Waaa.
    unwrapped = removeAllProxies(obj)

    unwrapped, wrapperCount = unwrapMethod(unwrapped)

    code = unwrapped.func_code
    defaults = unwrapped.func_defaults
    names = code.co_varnames[wrapperCount:code.co_argcount]

    nargs = len(names)
    if not positional:
        args = []
    else:
        args = list(positional)
        if len(args) > nargs:
            given = len(args)
            if wrapperCount:
                given += wrapperCount
            raise TypeError('%s() takes at most %d argument%s(%d given)' % (
                getattr(unwrapped, '__name__', repr(obj)),
                code.co_argcount,
                (code.co_argcount > 1 and 's ' or ' '),
                given))

    get = request.get
    nrequired = len(names)
    if defaults:
        nrequired -= len(defaults)

    for index in range(len(args), nargs):
        name = names[index]
        v = get(name, _marker)
        if v is _marker:
            if name == 'REQUEST':
                v = request
            elif index < nrequired:
                raise TypeError('Missing argument to %s(): %s' % (
                    getattr(unwrapped, '__name__', repr(obj)), name))
            else:
                v = defaults[index - nrequired]
        args.append(v)

    args = tuple(args)

    if __debug__:
        return debug_call(obj, args)

    return obj(*args)

def debug_call(obj, args):
    # The presence of this function allows us to set a pdb breakpoint
    obj = emergency_hack_2010_12_17(obj, 'within debug_call')
    return obj(*args)

def publish(request, handle_errors=True):
    try: # finally to clean up to_raise and close request
        to_raise = None
        while True:
            publication = request.publication
            try:
                try:
                    obj = None
                    try:
                        try:
                            request.processInputs()
                            publication.beforeTraversal(request)

                            obj = publication.getApplication(request)
                            obj = emergency_hack_2010_12_17(
                                obj, 'after getApplication')
                            obj = request.traverse(obj)
                            obj = emergency_hack_2010_12_17(
                                obj, 'after traverse')
                            publication.afterTraversal(request, obj)
                            obj = emergency_hack_2010_12_17(
                                obj, 'after afterTraversal')

                            result = publication.callObject(request, obj)
                            response = request.response
                            if result is not response:
                                response.setResult(result)

                            publication.afterCall(request, obj)

                        except:
                            exc_info = sys.exc_info()
                            publication.handleException(
                                obj, request, exc_info, True)

                            if not handle_errors:
                                # Reraise only if there is no adapter
                                # indicating that we shouldn't
                                reraise = component.queryAdapter(
                                    exc_info[1], IReRaiseException,
                                    default=None)
                                if reraise is None or reraise():
                                    raise
                    finally:
                        publication.endRequest(request, obj)

                    break # Successful.

                except Retry, retryException:
                    if request.supportsRetry():
                        # Create a copy of the request and use it.
                        newrequest = request.retry()
                        request.close()
                        request = newrequest
                    elif handle_errors:
                        # Output the original exception.
                        publication = request.publication
                        publication.handleException(
                            obj, request,
                            retryException.getOriginalException(), False)
                        break
                    else:
                        raise

            except:
                # Bad exception handler or retry method.
                # Re-raise after outputting the response.
                if handle_errors:
                    request.response.internalError()
                    to_raise = sys.exc_info()
                    break
                else:
                    raise

        response = request.response
        if to_raise is not None:
            raise to_raise[0], to_raise[1], to_raise[2]

    finally:
        to_raise = None  # Avoid circ. ref.
        request.close()  # Close database connections, etc.

    # Return the request, since it might be a different object than the one
    # that was passed in.
    return request


class DoNotReRaiseException(object):
    """Marker adapter for exceptions that should not be re-raised"""

    implements(IReRaiseException)

    def __init__(self, exc):
        pass

    def __call__(self):
        return False
